from typing import Annotated, TypeVar, Any
from contextlib import asynccontextmanager
from importlib.metadata import entry_points
import logging

from pydantic import BaseModel, HttpUrl, PlainSerializer
from fastapi import FastAPI, Depends, BackgroundTasks, HTTPException, status
from fastapi.middleware.cors import CORSMiddleware

from hatchet_sdk import EmptyModel, WorkflowConfig

# from hatchet_sdk.runnables.workflow import Workflow
from jwt import decode as decode_jwt

from knoc import __version__
from knoc.settings import settings
from knoc.workflows import all_wfs
from knoc.workflows.utils import ImportReport
from knoc.db import get_session

logger = logging.getLogger("knoc.server")


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Load the FastAPI app
    logger.debug("App is Starting Up!")
    yield
    logger.info(
        "App is Shutting Down Gracefully!",
        extra={"markup": True},
    )


app = FastAPI(lifespan=lifespan)
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.server.cors,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def get_plugins():
    for entry_point in entry_points().select(group="knoc.api"):
        try:
            api = entry_point.load()
        except Exception as err:
            logger.warning(
                f"Cannot load settings module {entry_point.name} - see below"
            )
            logger.exception(err)
            continue
        if not isinstance(api, FastAPI):
            logger.warning(f"Settings module {entry_point.name} is not a FastAPI")
            continue
        app.mount("/" + entry_point.name, api)
        logger.info(f"Mounted {entry_point.name}")


get_plugins()


class ServerInfo(BaseModel):
    version: str
    hatchet_url: HttpUrl
    api_doc: str


@app.get("/")
async def root() -> ServerInfo:
    hatchet_console_url = decode_jwt(
        settings.hatchet_client.token, options={"verify_signature": False}
    )["server_url"]
    return ServerInfo(
        version=__version__,
        hatchet_url=hatchet_console_url,
        api_doc="/docs",
    )


def ser_type(value: type[Any]) -> str:
    return value.__name__


## Ability to serialize classes, required for handling directly Hatchet workflows
T = TypeVar("T")
JSONSerializableType = Annotated[
    type[T], PlainSerializer(ser_type, when_used="json-unless-none")
]


class SerilizableWorkflowConfig(WorkflowConfig):
    input_validator: JSONSerializableType[BaseModel] | EmptyModel = (  # type: ignore
        EmptyModel
    )


@app.get("/wf")
async def list_workflows() -> dict[str, SerilizableWorkflowConfig]:
    """List the workflows"""
    return {k: wf.config for k, wf in all_wfs.items()}  # type: ignore


@app.post("/wf/{workflow_id}")
async def run_workflow(workflow_id: str):
    """Execute a workflow"""
    wf = all_wfs[workflow_id]
    result = await wf.aio_run()
    return result
